/*
 * Copyright (C) 2007 NetXen Inc.
 * 
 * All rights reserved.
 * 
 * The code contained in this file has been developed by NetXen Inc.
 * 
 * By downloading, compiling or using the code in any form, you agree to be bound
 * by the following terms:
 * 
 * 1) You may compile the unmodified source code file and reproduce and install
 * the resulting binary code file solely for your internal use with NetXen Inc.
 * hardware and systems you have purchased.
 * 
 * 2) Except as expressly permitted herein or in a separately signed agreement,
 * any other use, reproduction, distribution, modification, display, creation
 * of derivative works or circulation of the code is strictly prohibited.  You
 * may not transfer the code or assign your rights or obligations hereunder
 * without the prior written consent of NetXen Inc.
 * 
 * 3) This file contains confidential, proprietary and trade secret information
 * of NetXen Inc. and may not be disclosed to any third party or used for any
 * other purpose, except as expressly permitted herein.
 * 
 * 4) THIS SOFTWARE IS PROVIDED “AS IS” AND ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL NETXEN INC.
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.  THE APPLICATION OF THE UNITED NATIONS CONVENTION
 * ON CONTRACTS FOR THE INTERNATIONAL SALE OF GOODS IS HEREBY DISCLAIMED.
 * 
 * If you do not agree to the foregoing terms, you may not download, use or
 * compile this code in any form and must promptly delete or destroy all copiesi
 * of the code you have downloaded, compiled or reproduced.
 * 
 * Contact Information:
 * licensing@netxen.com
 * NetXen Inc.
 * 3965 Freedom Circle, Fourth Floor
 * Santa Clara, CA 95054
 * 
 */
/*
 * Copyright and CVS version tags.
 */

/*
 * This has the global functions for the corresponding 'c' file.
 */

#ifndef	_FLOW_TABLE_H
#define	_FLOW_TABLE_H

#include <linux/types.h>
#include <linux/sched.h>
#include <asm/current.h>
#include <linux/interrupt.h>
#include <linux/spinlock.h>


#define FLOW_TABLE_SIZE				0x2000

#if defined(CONFIG_X86_64)
#define	UNM_MAX_TCP_CONNS                       7500 
#else
#define UNM_MAX_TCP_CONNS                       1000
#endif

#define	UNM_INVALID_FLOW_ID                     ((uint32_t)0xffffffff)
#define UNM_INAVLID_FLOW_IDX			((uint32_t)0xffffffff)
#define UNM_INVALID_CONNID			((uint32_t)0xffffffff)

struct pegnet_sock_id_s;

/*
 * The flow ID struct - now a simple uint32_t
 */
typedef union {
	uint32_t	idx;
} flow_id_t;

/*
 * The free ID list structure.
 */
typedef struct {
	flow_id_t	*ids;
	uint32_t	max;
	uint32_t	head;
	uint32_t	tail;
} flow_free_ids_t;

/*
 * The flow element structure.
 */
typedef struct {
	flow_id_t	id;
	void		*flow;
	uint32_t	gnrtn; /* generation */
	atomic_t	refcnt;
} flow_tbl_entry_t;

/*
 * The flow table structure.
 */
typedef struct {
	flow_tbl_entry_t	*flows;
	uint32_t		max;
	flow_free_ids_t         free_list; /* this table tracks indices into
					      the flows array */
	int32_t			*connID_table; /* this array is used to index
						  with card-generated connID
						  and locate a flow */
	spinlock_t		lock;
	int nr_flows;
} flow_table_t;


/*
 * This function is used to allocate space for the flow table. It allocates the
 * maximum IDs and flow pointers. Does not allocate the flow itself.
 *
 * Parameters:
 *	None.
 *
 * Return:
 *	0		- If allocated successfully.
 *	ENOMEM		- If out of memory.
 */
int host_flowtbl_init(flow_table_t *tcp_tbl);

/*
 * This function is used to free space allocated for the flow table.
 *
 * Parameters:
 *	None.
 *
 * Return:
 *	None.
 */
void host_flowtbl_destroy(flow_table_t *tcp_tbl,void (*free_flow_cb)(flow_id_t id, void *flow));

/*
 * This function is used to allocate a flow from the flow table.
 * Returns a flow if allocated else returns NULL.
 * Also fills in the connID into the passed in parameter.
 *
 * Parameters:
 *	flow_idx: Index of flow in host's flow-id table	
 *
 * Return:
 *	The flow that was allocated.
 *	NULL	- If allocation fails.
 */
void *host_flowtbl_allocate_flow(flow_table_t *tcp_tbl,uint32_t *flow_idx);

/*
 * This function is used to free a flow from the table
 *
 *
 * Parameters:
 *	id		- The ID of the flow to free.
 *
 * Return:
 *	0		- If released successfully.
 *	-1		- If the flow is not found.
 */
int host_flowtbl_free_flow(flow_table_t *tcp_tbl,uint32_t connId);

int host_flowtbl_free_flow_by_flowidx(flow_table_t *tcp_tbl,uint32_t flow_idx);
void host_flowtbl_get_reference(flow_table_t *tcp_tbl,uint32_t id);

/*
 * This function is used to update the connID table with
 * the card-allocated connID
 *
 * Parameters:
 *	connID		- connID assigned by card
 *	sid		- sid that will use the connID
 *
 * Return:
 */
void host_flowtbl_set_flowid(uint32_t connID, struct pegnet_sock_id_s *sid);


/*
 * This function is used to get a flow from the flow table.
 *
 * Parameters:
 *	connId		- The ID of the flow to get.
 *
 * Return:
 *	The flow that was found.
 *	NULL		- If there is no flow for the ID.
 */
void *host_flowtbl_get_flow(flow_table_t *tcp_tbl,uint32_t connId);

/*
 * This function is used to get a flow from the flow table.
 *
 * Parameters:
 *	id		- The ID of the flow to get.
 *
 * Return:
 *	The flow that was found.
 *	NULL		- If there is no flow for the ID.
 */
void host_flowtbl_put_reference(flow_table_t *tcp_tbl,uint32_t id);

/*
 * This function is used to get the table index of a connection ID.
 *
 * Parameters:
 *	id		- The ID to convert to index.
 *
 * Return:
 *	The index into the table.
 */
uint16_t host_flowtbl_get_index(flow_table_t *tcp_tbl,uint32_t id);

/*
 * This function is used to get a flow from the flow table.
 *
 * Parameters:
 *	idx		- The index of the flow to get.
 *
 * Return:
 *	The flow that was found.
 *	NULL		- If there is no flow for the index.
 */
void *host_flowtbl_get_flow_by_idx(flow_table_t *tcp_tbl,uint16_t idx);

int host_flowtbl_get_nr_flows(flow_table_t *tcp_tbl);

/*
 * Temp context table for creating sk's on card
 */

#define UNM_INVALID_TMP_CTX		((uint32_t)0xffffffff)

typedef struct {
	struct list_head	list_entry;
	uint32_t		ctx;
	struct pegnet_sock_id_s	*sid;
} tmp_ctx_t;

typedef struct {
	struct list_head	free_list;
	struct list_head	used_list;
	unsigned int		num_free;
	spinlock_t              lock;
} tmp_ctx_table_t;

struct pegnet_sock_id_s *host_map_tmpctx_2_sid(flow_table_t *tcp_tbl,uint32_t tmp_ctx);
void host_update_connID_table(flow_table_t *tcp_tbl,
				struct pegnet_sock_id_s *sid, uint32_t connId);
void host_reset_connID_table(flow_table_t *tcp_tbl,
				struct pegnet_sock_id_s *sid,uint32_t connId);

#endif	/* _FLOW_TABLE_H */
