/*
 * Copyright (C) 2007 NetXen Inc.
 * 
 * All rights reserved.
 * 
 * The code contained in this file has been developed by NetXen Inc.
 * 
 * By downloading, compiling or using the code in any form, you agree to be bound
 * by the following terms:
 * 
 * 1) You may compile the unmodified source code file and reproduce and install
 * the resulting binary code file solely for your internal use with NetXen Inc.
 * hardware and systems you have purchased.
 * 
 * 2) Except as expressly permitted herein or in a separately signed agreement,
 * any other use, reproduction, distribution, modification, display, creation
 * of derivative works or circulation of the code is strictly prohibited.  You
 * may not transfer the code or assign your rights or obligations hereunder
 * without the prior written consent of NetXen Inc.
 * 
 * 3) This file contains confidential, proprietary and trade secret information
 * of NetXen Inc. and may not be disclosed to any third party or used for any
 * other purpose, except as expressly permitted herein.
 * 
 * 4) THIS SOFTWARE IS PROVIDED “AS IS” AND ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL NETXEN INC.
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.  THE APPLICATION OF THE UNITED NATIONS CONVENTION
 * ON CONTRACTS FOR THE INTERNATIONAL SALE OF GOODS IS HEREBY DISCLAIMED.
 * 
 * If you do not agree to the foregoing terms, you may not download, use or
 * compile this code in any form and must promptly delete or destroy all copiesi
 * of the code you have downloaded, compiled or reproduced.
 * 
 * Contact Information:
 * licensing@netxen.com
 * NetXen Inc.
 * 3965 Freedom Circle, Fourth Floor
 * Santa Clara, CA 95054
 * 
 */

/* 
 * This is the intercept module that will arbitrate between offload and 
 * native stack. Since offload decision is taken in listen() and connect(),
 * they are unconditionally passed on to offload stack. For other calls, a 
 * check for offload is made.
 */
#include <linux/module.h>
#include <linux/version.h>
#include <asm/atomic.h>
#include <linux/net.h>
#include <net/sock.h>
#include <linux/poll.h>
#include "unm_host_pnic_types.h"

extern struct proto_ops inet_stream_ops;

typedef struct {
	atomic_t ref_count;
	struct proto_ops inet_stream_ops;
	struct proto_ops host_inet_stream_ops;
	int is_registered;
} nx_tnic_intercept_t;

static nx_tnic_intercept_t nx_tnic_intercept;

int nx_intercept_inet_release(struct socket *sock)
{
	int rv;
	
	if (UNM_SOCK_IS_SOCK_OFFLOADED(sock)) {
		rv = nx_tnic_intercept.inet_stream_ops.release(sock);
	} else {
		rv =  nx_tnic_intercept.host_inet_stream_ops.release(sock);
	}

	return rv;
		
}

int nx_intercept_inet_socketpair(struct socket *sock1, struct socket *sock2)
{
	int rv;
	
	if (UNM_SOCK_IS_SOCK_OFFLOADED(sock1)) {
		rv = nx_tnic_intercept.inet_stream_ops.socketpair(sock1, sock2);
	} else {
		rv =  nx_tnic_intercept.host_inet_stream_ops.socketpair(sock1, sock2);
	}

	return rv;
	
}

int nx_intercept_inet_accept(struct socket *sock, struct socket *newsock, int flags)
{
	int rv;
	
	if (UNM_SOCK_IS_SOCK_OFFLOADED(sock)) {
		rv = nx_tnic_intercept.inet_stream_ops.accept(sock, newsock, flags);
	} else {
		rv =  nx_tnic_intercept.host_inet_stream_ops.accept(sock, newsock, flags);
		if (nx_tnic_intercept.is_registered && newsock->sk) { 
			newsock->sk->SK_PROTINFO = NULL;
		}
	}

	return rv;
	
}

int nx_intercept_inet_getname(struct socket *sock, struct sockaddr *uaddr,
				int *uaddr_len, int peer)
{
	int rv;
	
	if (UNM_SOCK_IS_SOCK_OFFLOADED(sock)) {
		rv = nx_tnic_intercept.inet_stream_ops.getname(sock, uaddr, 
								uaddr_len, peer);
	} else {
		rv =  nx_tnic_intercept.host_inet_stream_ops.getname(sock, uaddr,
								uaddr_len, peer);
	}

	return rv;
	
}

unsigned int nx_intercept_inet_poll(struct file *file, struct socket *sock,
					 poll_table *wait)
{
	int rv;
	
	if (UNM_SOCK_IS_SOCK_OFFLOADED(sock)) {
		rv = nx_tnic_intercept.inet_stream_ops.poll(file, sock, wait);
	} else {
		rv =  nx_tnic_intercept.host_inet_stream_ops.poll(file, sock, wait);
	}

	return rv;
	
}

int nx_intercept_inet_ioctl(struct socket *sock, unsigned int cmd, unsigned long arg)
{
	int rv;
	
	if (UNM_SOCK_IS_SOCK_OFFLOADED(sock)) {
		rv = nx_tnic_intercept.inet_stream_ops.ioctl(sock, cmd, arg);
	} else {
		rv =  nx_tnic_intercept.host_inet_stream_ops.ioctl(sock, cmd, arg);
	}

	return rv;
	
}

int nx_intercept_inet_shutdown(struct socket *sock, int how)
{
	int rv;
	
	if (UNM_SOCK_IS_SOCK_OFFLOADED(sock)) {
		rv = nx_tnic_intercept.inet_stream_ops.shutdown(sock, how);
	} else {
		rv =  nx_tnic_intercept.host_inet_stream_ops.shutdown(sock, how);
	}

	return rv;
	
}

int nx_intercept_inet_setsockopt(struct socket *sock, int level, int optname,
					 char *optval, int optlen)
{
	int rv;
	
	if (UNM_SOCK_IS_SOCK_OFFLOADED(sock)) {
		rv = nx_tnic_intercept.inet_stream_ops.setsockopt(sock, level, optname, 
								optval, optlen);
	} else {
		rv =  nx_tnic_intercept.host_inet_stream_ops.setsockopt(sock, level, optname,
								optval, optlen);
	}

	return rv;
	
}

int nx_intercept_inet_getsockopt(struct socket *sock, int level, int optname,
					char *optval, int *optlen)
{
	int rv;
	
	if (UNM_SOCK_IS_SOCK_OFFLOADED(sock)) {
		rv = nx_tnic_intercept.inet_stream_ops.getsockopt(sock, level, optname, 
								optval, optlen);
	} else {
		rv =  nx_tnic_intercept.host_inet_stream_ops.getsockopt(sock, level, optname,
								optval, optlen);
	}

	return rv;
}

int nx_intercept_inet_sendmsg(struct kiocb *iocb, struct socket *sock,
				struct msghdr *msg, size_t size)
{
	int rv;
	
	if (UNM_SOCK_IS_SOCK_OFFLOADED(sock)) {
		rv = nx_tnic_intercept.inet_stream_ops.sendmsg(iocb, sock, msg, size);
	} else {
		rv =  nx_tnic_intercept.host_inet_stream_ops.sendmsg(iocb, sock, msg, size);
	}

	return rv;
	
}

int nx_intercept_inet_recvmsg(struct kiocb *iocb, struct socket *sock,
				struct msghdr *msg, size_t size, int flags)
{
	int rv;
	
	if (UNM_SOCK_IS_SOCK_OFFLOADED(sock)) {
		rv = nx_tnic_intercept.inet_stream_ops.recvmsg(iocb, sock, msg, size, flags);
	} else {
		rv =  nx_tnic_intercept.host_inet_stream_ops.recvmsg(iocb, sock, msg, size, flags);
	}

	return rv;
	
}

ssize_t nx_intercept_inet_sendpage(struct socket *sock, struct page *page, int offset,
					size_t size, int flags)
{
	int rv;
	
	if (UNM_SOCK_IS_SOCK_OFFLOADED(sock)) {
		rv = nx_tnic_intercept.inet_stream_ops.sendpage(sock, page, offset, size, flags);
	} else {
		rv =  nx_tnic_intercept.host_inet_stream_ops.sendpage(sock, page, offset, size, flags);
	}

	return rv;
	
}

int nx_intercept_inet_listen(struct socket *sock, int backlog)
{
	int rv;

	if (nx_tnic_intercept.is_registered) {
		atomic_inc(&nx_tnic_intercept.ref_count);
		if (nx_tnic_intercept.is_registered) {
			 rv = nx_tnic_intercept.inet_stream_ops.listen(sock, backlog);
		} else {
			atomic_dec(&nx_tnic_intercept.ref_count);
			goto no_tnic;
		}
		atomic_dec(&nx_tnic_intercept.ref_count);
	} else {
no_tnic:
		 rv =  nx_tnic_intercept.host_inet_stream_ops.listen(sock, backlog);
	}

	return rv;
}

int nx_intercept_inet_connect(struct socket *sock, struct sockaddr *uaddr,
				 int addr_len, int flags)
{
	int rv;

	if (nx_tnic_intercept.is_registered) {
		atomic_inc(&nx_tnic_intercept.ref_count);
		if (nx_tnic_intercept.is_registered) {
			 rv = nx_tnic_intercept.inet_stream_ops.connect(sock, uaddr, addr_len, flags);
		} else {
			atomic_dec(&nx_tnic_intercept.ref_count);
			goto no_tnic;
		}
		atomic_dec(&nx_tnic_intercept.ref_count);
	} else {
no_tnic:
		 rv =  nx_tnic_intercept.host_inet_stream_ops.connect(sock, uaddr, addr_len, flags);
	}

	return rv;

}

int nx_tnic_register_ops(struct proto_ops *inet_ops, struct proto_ops *host_ops)
{
	if (nx_tnic_intercept.is_registered)
		return -EINVAL;

	nx_tnic_intercept.inet_stream_ops.release 	= inet_ops->release;
	nx_tnic_intercept.inet_stream_ops.socketpair	= inet_ops->socketpair;
	nx_tnic_intercept.inet_stream_ops.accept	= inet_ops->accept;
	nx_tnic_intercept.inet_stream_ops.getname	= inet_ops->getname;
	nx_tnic_intercept.inet_stream_ops.poll		= inet_ops->poll;
	nx_tnic_intercept.inet_stream_ops.ioctl		= inet_ops->ioctl;
	nx_tnic_intercept.inet_stream_ops.shutdown	= inet_ops->shutdown;
	nx_tnic_intercept.inet_stream_ops.setsockopt	= inet_ops->setsockopt;
	nx_tnic_intercept.inet_stream_ops.getsockopt	= inet_ops->getsockopt;
	nx_tnic_intercept.inet_stream_ops.sendmsg	= inet_ops->sendmsg;
	nx_tnic_intercept.inet_stream_ops.recvmsg	= inet_ops->recvmsg;
	nx_tnic_intercept.inet_stream_ops.sendpage	= inet_ops->sendpage;
	nx_tnic_intercept.inet_stream_ops.listen	= inet_ops->listen;
	nx_tnic_intercept.inet_stream_ops.connect	= inet_ops->connect;


	memcpy(host_ops,  &nx_tnic_intercept.host_inet_stream_ops, sizeof(*host_ops));
	nx_tnic_intercept.is_registered = 1;

	return 0;
}

EXPORT_SYMBOL(nx_tnic_register_ops);

void nx_tnic_unregister_ops(void)
{
	nx_tnic_intercept.is_registered = 0;

retry:
	/* wait for refcount to become 0 */
	if (atomic_read(&nx_tnic_intercept.ref_count)) {

		printk("ref count not 0 :%d\n", atomic_read(&nx_tnic_intercept.ref_count));
		current->state = TASK_UNINTERRUPTIBLE;
		schedule_timeout(3);
		current->state = TASK_RUNNING;
		goto retry;
	}

	/* all done */
	memset(&nx_tnic_intercept.inet_stream_ops, 0, sizeof(nx_tnic_intercept.inet_stream_ops));
}

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,16)  || \
    (LINUX_VERSION_CODE == KERNEL_VERSION(2,6,15) && (FEDORA==FC5))) && \
    defined(CONFIG_DEBUG_RODATA)
static int
change_perm(unsigned long ptr, unsigned int size, pgprot_t prot)
{
        unsigned long endptr = (ptr + size + PAGE_SIZE - 1) & PAGE_MASK;
        size = (size + PAGE_SIZE - 1) >> PAGE_SHIFT;
        for (ptr &= PAGE_MASK; ptr < endptr; ptr += PAGE_SIZE) {
               if (change_page_attr(virt_to_page(ptr), 1, prot) < 0) {
                       return -1;
               }
        }
        global_flush_tlb();
        return 0;
}
#endif

EXPORT_SYMBOL(nx_tnic_unregister_ops);

static int __init nx_intercept_init(void)
{
#if 1
	if (!try_module_get(THIS_MODULE))
		return -EINVAL;
#endif
	nx_tnic_intercept.is_registered	= 0;

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,16)  || \
    (LINUX_VERSION_CODE == KERNEL_VERSION(2,6,15) && (FEDORA==FC5))) && \
    defined(CONFIG_DEBUG_RODATA)
        if (change_perm((unsigned long)&inet_stream_ops,
                         sizeof(inet_stream_ops),
                         PAGE_KERNEL_FLAG) < 0) {
		module_put(THIS_MODULE);
                return -EIO;
        }
#endif

	memcpy(&nx_tnic_intercept.host_inet_stream_ops, &inet_stream_ops,
		sizeof(inet_stream_ops));

        inet_stream_ops.release = nx_intercept_inet_release;
        inet_stream_ops.connect = nx_intercept_inet_connect;
        inet_stream_ops.socketpair = nx_intercept_inet_socketpair;
        inet_stream_ops.accept = nx_intercept_inet_accept;
        inet_stream_ops.getname = nx_intercept_inet_getname;
        inet_stream_ops.poll = nx_intercept_inet_poll;
        inet_stream_ops.ioctl = nx_intercept_inet_ioctl;
        inet_stream_ops.listen = nx_intercept_inet_listen;
        inet_stream_ops.shutdown = nx_intercept_inet_shutdown;
        inet_stream_ops.setsockopt = nx_intercept_inet_setsockopt;
        inet_stream_ops.getsockopt = nx_intercept_inet_getsockopt;
        inet_stream_ops.sendmsg = nx_intercept_inet_sendmsg;
        inet_stream_ops.recvmsg = nx_intercept_inet_recvmsg;
        inet_stream_ops.sendpage = nx_intercept_inet_sendpage;

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,16)  || \
    (LINUX_VERSION_CODE == KERNEL_VERSION(2,6,15) && (FEDORA==FC5))) && \
    defined(CONFIG_DEBUG_RODATA)
        change_perm((unsigned long)&inet_stream_ops, sizeof(inet_stream_ops),
                    PAGE_KERNEL_RO);
#endif

	atomic_set(&nx_tnic_intercept.ref_count, 0);
	memset(&nx_tnic_intercept.inet_stream_ops, 0, sizeof(nx_tnic_intercept.inet_stream_ops));

	return 0;
}

static void __exit nx_intercept_exit(void)
{
	/* this isn't allowed */
	BUG();

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,16)  || \
    (LINUX_VERSION_CODE == KERNEL_VERSION(2,6,15) && (FEDORA==FC5))) && \
    defined(CONFIG_DEBUG_RODATA)
        if (change_perm((unsigned long)&inet_stream_ops,
                         sizeof(inet_stream_ops),
                         PAGE_KERNEL_FLAG) < 0) {
		return;
        }
#endif

	memcpy(&inet_stream_ops, &nx_tnic_intercept.host_inet_stream_ops,
		sizeof(inet_stream_ops));

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,16)  || \
    (LINUX_VERSION_CODE == KERNEL_VERSION(2,6,15) && (FEDORA==FC5))) && \
    defined(CONFIG_DEBUG_RODATA)
        change_perm((unsigned long)&inet_stream_ops, sizeof(inet_stream_ops),
                    PAGE_KERNEL_RO);
#endif

}

module_init(nx_intercept_init);
module_exit(nx_intercept_exit);

