/*
 * Copyright (C) 2007 NetXen Inc.
 * 
 * All rights reserved.
 * 
 * The code contained in this file has been developed by NetXen Inc.
 * 
 * By downloading, compiling or using the code in any form, you agree to be bound
 * by the following terms:
 * 
 * 1) You may compile the unmodified source code file and reproduce and install
 * the resulting binary code file solely for your internal use with NetXen Inc.
 * hardware and systems you have purchased.
 * 
 * 2) Except as expressly permitted herein or in a separately signed agreement,
 * any other use, reproduction, distribution, modification, display, creation
 * of derivative works or circulation of the code is strictly prohibited.  You
 * may not transfer the code or assign your rights or obligations hereunder
 * without the prior written consent of NetXen Inc.
 * 
 * 3) This file contains confidential, proprietary and trade secret information
 * of NetXen Inc. and may not be disclosed to any third party or used for any
 * other purpose, except as expressly permitted herein.
 * 
 * 4) THIS SOFTWARE IS PROVIDED “AS IS” AND ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL NETXEN INC.
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.  THE APPLICATION OF THE UNITED NATIONS CONVENTION
 * ON CONTRACTS FOR THE INTERNATIONAL SALE OF GOODS IS HEREBY DISCLAIMED.
 * 
 * If you do not agree to the foregoing terms, you may not download, use or
 * compile this code in any form and must promptly delete or destroy all copiesi
 * of the code you have downloaded, compiled or reproduced.
 * 
 * Contact Information:
 * licensing@netxen.com
 * NetXen Inc.
 * 3965 Freedom Circle, Fourth Floor
 * Santa Clara, CA 95054
 * 
 */
/*
 * NetXen:
 *
 * Provides a cache for destination lookup.
 */

#include <linux/net.h>
#include <linux/netdevice.h>
#include <net/dst.h>
#include <net/route.h>
#include <linux/if_vlan.h>
#include <linux/inetdevice.h>
#include <linux/if.h>
#include <linux/netlink.h>
#include <linux/rtnetlink.h>
#include <linux/socket.h>

#include "unm_host_pnic_types.h"
#include "nx_route_table.h"
#include "unm_host_pnic_types.h"

#ifndef sock_create_kern
#define sock_create_kern        sock_create
#endif

static int nx_route_listener_init(void);
static void nx_route_listener_destroy(void);
static void nx_rt_sync_queue_proc(void *data);
static void nx_rt_remove_sync_queue(nx_tnic_adapter_t *adapter);

static nx_host_rt_sync_work_queue_t	sync_q;	/* Adapters add them self
						   into this list for
						   synchronizing */
/*
 *
 */
int nx_route_tbl_init(void)
{
	INIT_NX_HBUCKET_HEAD(&sync_q.list);
	NX_KC_INIT_WORK(&sync_q.task, nx_rt_sync_queue_proc, NULL);
	nx_route_listener_init();
	return (0);
}

/*
 *
 */
void nx_route_tbl_destroy(void)
{
	nx_route_listener_destroy();
}

/*
 *
 */
int nx_route_tbl_create(nx_tnic_adapter_t *adapter)
{
	int			i;
	nx_host_rt_entry_t	*rt;

	nx_print("Creating route table\n");

	adapter->rt_table =
		(nx_host_route_tbl_t *)vmalloc(sizeof (nx_host_route_tbl_t));
	if (adapter->rt_table == NULL) {
		nx_print("Route table alloc failed\n");
		return (-ENOMEM);
	}

	for (i = 0; i < RC_HASH_TBL_SIZE; i++) {
		INIT_NX_HBUCKET_HEAD(&adapter->rt_table->routes[i]);
	}

	adapter->rt_table->cnt = 0;
	INIT_LIST_HEAD(&adapter->rt_table->free_list);
	spin_lock_init(&adapter->rt_table->lock.lock);
	adapter->rt_table->lock.inuse = 0;
	init_waitqueue_head(&adapter->rt_table->lock.wq);

	for (i = 0; i < RC_MAX_ENTRIES; i++) {
		rt = (nx_host_rt_entry_t *)kmalloc(sizeof (nx_host_rt_entry_t),
						   GFP_KERNEL);
		if (rt == NULL) {
			nx_print("Memory alloc for route entries failed\n");
			nx_route_tbl_free(adapter);
			return (-ENOMEM);
		}
		list_add(&rt->list, &adapter->rt_table->free_list);
	}

	return (0);
}

/*
 * 
 */
static void lock_rt_table(nx_host_route_tbl_t *table)
{
        might_sleep();

        spin_lock_bh(&table->lock.lock);

        if (table->lock.inuse) {
		DEFINE_WAIT(wait);

		for(;;) {
			prepare_to_wait_exclusive(&table->lock.wq, &wait,
						  TASK_UNINTERRUPTIBLE);
			spin_unlock_bh(&table->lock.lock);
			schedule();
			spin_lock_bh(&table->lock.lock);
			if(!table->lock.inuse) {
				break;
			}
		}
		finish_wait(&table->lock.wq, &wait);
	}
        table->lock.inuse = 1;
        spin_unlock_bh(&table->lock.lock);
}

/*
 *
 */
static void unlock_rt_table(nx_host_route_tbl_t *table)
{
        spin_lock_bh(&table->lock.lock);
        table->lock.inuse = 0;
        if (waitqueue_active(&table->lock.wq)) {
                wake_up(&table->lock.wq);
	}
        spin_unlock_bh(&table->lock.lock);
}

/*
 *
 */
void nx_route_tbl_free(nx_tnic_adapter_t *adapter)
{
	int			i;
	nx_host_rt_entry_t	*rt;
	nx_hbucket_head_t	*head;
	struct list_head 	*node1;
	struct list_head	*node2;

	if (adapter->rt_table == NULL) {
		return;
	}

	lock_rt_table(adapter->rt_table);

	list_for_each_safe(node1, node2, &adapter->rt_table->free_list) {
		rt = list_entry(node1, nx_host_rt_entry_t, list);
		list_del(&rt->list);
		kfree(rt);
	}

	for (i = 0; i < RC_HASH_TBL_SIZE; i++) {
		head = &adapter->rt_table->routes[i];

		spin_lock_bh(&head->lock);

		list_for_each_safe(node1, node2, &head->head) {
			rt = list_entry(node1, nx_host_rt_entry_t, list);
			list_del(&rt->list);
			kfree(rt);
		}

		spin_unlock_bh(&head->lock);
	}

	unlock_rt_table(adapter->rt_table);

	nx_rt_remove_sync_queue(adapter);

	vfree(adapter->rt_table);

	adapter->rt_table = NULL;
}

/*
 *
 */
static nx_host_rt_entry_t *nx_route_alloc(nx_host_route_tbl_t *table)
{
	nx_host_rt_entry_t	*rt;

        spin_lock_bh(&table->lock.lock);

	if (!list_empty(&table->free_list)) {
		rt = list_entry(table->free_list.next, nx_host_rt_entry_t,
				list);
		list_del_init(&rt->list);
		table->cnt++;
	} else {
		nx_print("Memory alloc for route entries failed\n");
		rt = NULL;
	}

        spin_unlock_bh(&table->lock.lock);

	if (rt) {
		memset(rt, 0, sizeof (nx_host_rt_entry_t));
	}

	return (rt);
}

/*
 *
 */
static inline void nx_route_free(nx_host_route_tbl_t *table,
				 nx_host_rt_entry_t *rt)
{
	if (rt->refcnt > 0) {
		nx_print("Freeing a route with refcnt[%d] > 0\n", rt->refcnt);
	}

        spin_lock_bh(&table->lock.lock);

	list_add_tail(&rt->list, &table->free_list);
	table->cnt--;
	if (table->cnt < 0) {
		nx_print("Too many frees\n");
	}

        spin_unlock_bh(&table->lock.lock);
}

/*
 *
 */
static inline nx_hbucket_head_t *get_hbucket_head(nx_host_route_tbl_t *table,
						  uint32_t saddr,
						  uint32_t daddr)
{
	uint32_t	hash;

	hash = ((daddr & 0xF0F0F0F0) >> 4) |
		((daddr & 0x0F0F0F0F) << 4);
	hash ^= saddr;
	hash ^= (hash >> 16);
	hash = (hash ^ (hash >> 8)) & (RC_HASH_TBL_SIZE - 1);

	return (&table->routes[hash]);
}

/*
 *
 */
static void copy_route(struct rtable *req, uint8_t port,
		       nx_host_rt_entry_t *new)
{
	struct net_device	*dev;

	new->key.daddr = req->fl.nl_u.ip4_u.daddr;
	new->key.saddr = req->fl.nl_u.ip4_u.saddr;

	new->mtu = (NX_KC_DST_MTU(&req->u.dst) > 1500) ?
			1500 : NX_KC_DST_MTU(&req->u.dst);
	new->window = 0;
	new->irtt = 0;
	new->advmss = (dst_metric(&req->u.dst, RTAX_ADVMSS) > 1460) ?
			1460 : dst_metric(&req->u.dst, RTAX_ADVMSS);

	new->nhs[0].gw = req->rt_gateway;
	new->nhs[0].tos = req->fl.nl_u.ip4_u.tos;

	new->nh_cnt = 1;

        dev = NX_KC_ROUTE_DEVICE(req);

	if (dev->priv_flags & IFF_802_1Q_VLAN) {
		new->nhs[0].vlanid = VLAN_DEV_INFO(dev)->vlan_id;
		new->nhs[0].vlan_set = 1;
		new->nhs[0].qos = 0/* VLAN_DEV_INFO(dev)->qos */;
	} else {
		new->nhs[0].vlanid = 0;
		new->nhs[0].vlan_set = 0;
		new->nhs[0].qos = 0;
	}

	new->nhs[0].port = port;
	new->invalid = 0;

	return;
}

/*
 *
 */
static inline int cmp_route(nx_host_rt_key_t *a1, nx_host_rt_key_t *a2)
{
	if (a1->daddr != a2->daddr) {
		return ((int)a1->daddr - (int)a2->daddr);
	}

	return ((int)a1->saddr - (int)a2->saddr);
}

/*
 *
 */
static inline void list_insert_before(struct list_head *new,
				      struct list_head *before)
{
	new->next = before;
	new->prev = before->prev;
	new->prev->next = new;
	before->prev = new;
}

/*
 *
 */
static inline void form_route_key(nx_host_rt_key_t *key, struct rtable *req)
{
	key->saddr = req->fl.nl_u.ip4_u.saddr;
	key->daddr = req->fl.nl_u.ip4_u.daddr;
}

/*
 *
 */
static void copy_to_msg(nx_host_rt_entry_t *entry, routecache_request_t *msg,
			uint8_t cmd)
{
	memset(msg, 0, sizeof (routecache_request_t));

	msg->req_type = cmd;
	msg->type = UNM_MSGTYPE_ROUTE_REQUEST;
	msg->src.v4 = entry->key.saddr;
	msg->dest.v4 = entry->key.daddr;
	msg->gateway.v4 = entry->nhs[0].gw;
	msg->advmss = entry->advmss;
	msg->pmtu = entry->mtu;
	msg->vlanid = entry->nhs[0].vlanid;
	msg->tos = entry->nhs[0].tos;
	msg->port = entry->nhs[0].port;
	msg->vlan_set = entry->nhs[0].vlan_set;
	msg->qos = entry->nhs[0].qos;
	msg->ip_v6 = 0;

	return;
}

/*
 *
 */
static int send_route_to_adapter(nx_tnic_adapter_t *adapter,
				 routecache_request_t *msg)
{
	int	cnt = 0;

	for (cnt = 0; cnt < NX_TNIC_MAX_PORTS; cnt++) {
		if (NX_GET_DEV_FROM_TNIC_ADAPTER(adapter,cnt)) {
			break;
		}
	}

	if (cnt >= NX_TNIC_MAX_PORTS) {
		nx_print("No device is up for tnic_adapter=%p\n", adapter);
		return -1;
	}
	return (unm_send_h2p_msg(NX_GET_DEV_FROM_TNIC_ADAPTER(adapter,cnt), 
				(unm_msg_t *)msg,sizeof (unm_msg_body_t)));
}

/*
 *
 */
int nx_route_add(nx_tnic_adapter_t *adapter, struct rtable *req, uint8_t port)
{
	nx_host_rt_entry_t	*entry = NULL;
	nx_host_rt_entry_t	*new;
	nx_hbucket_head_t	*head;
	nx_host_rt_key_t	key;
	struct list_head 	*node1;
	struct list_head	*node2;
	int			rv;
	int			new_entry = 0;
	routecache_request_t	msg;


	new = nx_route_alloc(adapter->rt_table);
	if (new == NULL) {
		nx_print("Cannot allocate route cache. Aborting\n");
		return (-ENOMEM);
	}

	form_route_key(&key, req);
	nx_print1("ADD: saddr[0x%x], daddr[0x%x]\n", key.saddr, key.daddr);

	head = get_hbucket_head(adapter->rt_table, key.saddr, key.daddr);

	spin_lock_bh(&head->lock);

	list_for_each_safe(node1, node2, &head->head) {
		entry = list_entry(node1, nx_host_rt_entry_t, list);
		rv = cmp_route(&key, &entry->key);
		if (rv == 0) {
			nx_print1("Route exists\n");
			entry->refcnt++;
			goto done;
		}
		if (rv < 0) {
			entry = NULL;
		}
	}

	copy_route(req, port, new);

	new->refcnt = 1;

	if (entry) {
		list_insert_before(&new->list, &entry->list);
	} else {
		list_add_tail(&new->list, &head->head);
	}
	rv = 0;
	copy_to_msg(new, &msg, NX_ROUTECACHE_ADD);
	new_entry = 1;

  done:
	spin_unlock_bh(&head->lock);

	if (new_entry) {
		rv = send_route_to_adapter(adapter, &msg);
		if (rv) {
			nx_print("Failed to send route msg to card\n");
			nx_route_remove(adapter, &key);
		}
	} else {
		nx_route_free(adapter->rt_table, new);
	}

	return (rv);
}

/*
 *
 */
int nx_route_update(nx_tnic_adapter_t *adapter, struct rtable *req,
		    uint8_t port)
{
	nx_host_rt_entry_t	*entry;
	nx_hbucket_head_t	*head;
	struct list_head 	*node1;
	struct list_head	*node2;
	int			rv = -EINVAL;
	nx_host_rt_key_t	key;
	routecache_request_t	msg;


	form_route_key(&key, req);
	nx_print1("Update: saddr[0x%x], daddr[0x%x]\n", key.saddr, key.daddr);

	head = get_hbucket_head(adapter->rt_table, key.saddr, key.daddr);

	spin_lock_bh(&head->lock);

	list_for_each_safe(node1, node2, &head->head) {
		entry = list_entry(node1, nx_host_rt_entry_t, list);
		rv = cmp_route(&key, &entry->key);
		if (rv == 0) {
			copy_route(req, port, entry);
			copy_to_msg(entry, &msg, NX_ROUTECACHE_MODIFY);
			goto done;
		}
	}
	nx_print1("Route cache not found for saddr[0x%x], daddr[0x%x]\n",
		  key.saddr, key.daddr);

  done:
	spin_unlock_bh(&head->lock);

	if (rv == 0) {
		rv = send_route_to_adapter(adapter, &msg);
		if (rv) {
			nx_print("Sending Route update to card failed\n");
		}
	}
	return (rv);
}

/*
 *
 */
int nx_route_remove(nx_tnic_adapter_t *adapter, nx_host_rt_key_t *key)
{
	nx_host_rt_entry_t	*entry = NULL;
	nx_hbucket_head_t	*head;
	struct list_head 	*node1;
	struct list_head	*node2;
	int			rv = -EINVAL;
	routecache_request_t	msg;


	nx_print1("REMOVE: saddr[0x%x], daddr[0x%x]\n",
		  key->saddr, key->daddr);

	head = get_hbucket_head(adapter->rt_table, key->saddr, key->daddr);

	spin_lock_bh(&head->lock);

	list_for_each_safe(node1, node2, &head->head) {
		entry = list_entry(node1, nx_host_rt_entry_t, list);
		rv = cmp_route(key, &entry->key);
		if (rv == 0) {
			entry->refcnt--;
			if (entry->refcnt == 0) {
				list_del(&entry->list);
				copy_to_msg(entry, &msg, NX_ROUTECACHE_REMOVE);
			} else {
				rv = -1;
			}
			goto done;
		}
	}

	nx_print1("Route entry not found\n");
	rv = -EINVAL;

  done:
	spin_unlock_bh(&head->lock);

	if (!rv) {
		nx_route_free(adapter->rt_table, entry);
		rv = send_route_to_adapter(adapter, &msg);
		if (rv) {
			nx_print("Route remove could not be sent to card\n");
		}
	}
	return (rv);
}

/*
 *
 */
static int is_nx_route_changed(struct rtable *req, uint8_t port,
			       nx_host_rt_entry_t *entry)
{
	struct net_device	*dev;
	int			vlan_set = 0;
	int			vlanid = 0;

	dev = NX_KC_ROUTE_DEVICE(req);

	if (dev->priv_flags & IFF_802_1Q_VLAN) {
		vlanid = VLAN_DEV_INFO(dev)->vlan_id;
		vlan_set = 1;
	}

	return (entry->invalid ||
		entry->key.daddr != req->fl.nl_u.ip4_u.daddr ||
		entry->key.saddr != req->fl.nl_u.ip4_u.saddr ||
		entry->mtu != NX_KC_DST_MTU(&req->u.dst) ||
		entry->advmss != dst_metric(&req->u.dst, RTAX_ADVMSS) ||
		entry->nhs[0].gw != req->rt_gateway ||
		entry->nhs[0].tos != req->fl.nl_u.ip4_u.tos ||
		entry->nhs[0].vlanid != vlanid ||
		entry->nhs[0].vlan_set != vlan_set  ||
		entry->nhs[0].port != port);
}

/*
 *
 */
static void nx_route_sync(nx_host_rt_entry_t *entry,
			  nx_tnic_adapter_t *adapter)
{
	struct rtable		*rt;
	struct net_device	*dev;
	int			port;
	int			rv;
	nx_tnic_adapter_t	*new_adapter;
	routecache_request_t	msg;
	struct flowi		fl =
		{ .oif = 0,
		  .nl_u = { .ip4_u = { .daddr = entry->key.daddr,
				       .saddr = entry->key.saddr,
				       .tos   = entry->nhs[0].tos } },
		  .proto = IPPROTO_TCP,
		  .uli_u = { .ports = { .sport = 0,
					.dport = 0} } };

	nx_print1("looking up route for saddr[0x%x], daddr[0x%x]\n",
		  entry->key.saddr, entry->key.daddr);

	rv = ip_route_output_key(&rt, &fl);
	if (rv) {
		nx_print1("Could not find route for saddr[0x%x], daddr[0x%x]: "
			  "ERROR[%d]\n",
			  entry->key.saddr, entry->key.daddr, rv);
		entry->invalid = 1;
		copy_to_msg(entry, &msg, NX_ROUTECACHE_INVALIDATE);
		rv = send_route_to_adapter(adapter, &msg);
		if (rv) {
			nx_print("Sending Route invalidate to card failed\n");
		}
		return;
	}

	nx_print1("Found route\n");
	dev = NX_KC_ROUTE_DEVICE(rt);

	/*
	 * Once there is multi card code we need to check if we are going
	 * to the same card.
	 */
	new_adapter = nx_get_dev_tnic_adapter(dev);
	port = nx_get_device_port(dev);
	if (new_adapter != NULL && new_adapter == adapter && port >= 0) {
		if (is_nx_route_changed(rt, port, entry)) {
			nx_print1("Updating route saddr[0x%x], daddr[0x%x]\n",
				  entry->key.saddr, entry->key.daddr);
			copy_route(rt, port, entry);
			copy_to_msg(entry, &msg, NX_ROUTECACHE_MODIFY);
			rv = send_route_to_adapter(adapter, &msg);
			if (rv) {
				nx_print("Sending Route update to card failed\n");
			}
		} else {
			nx_print1("No change in route\n");
		}
	} else {
		nx_print1("Cannot be routed through our device\n");
		entry->invalid = 1;
		copy_to_msg(entry, &msg, NX_ROUTECACHE_INVALIDATE);
		rv = send_route_to_adapter(adapter, &msg);
		if (rv) {
			nx_print("Sending Route invalidate to card failed\n");
		}
	}

	ip_rt_put(rt);

	return;
}

/*
 *
 */
void nx_route_sync_table(nx_tnic_adapter_t *adapter)
{
	int			i;
	nx_host_rt_entry_t	*rt;
	nx_hbucket_head_t	*head;
	struct list_head 	*node1;
	struct list_head	*node2;

	nx_print1("Inside\n");
	lock_rt_table(adapter->rt_table);

	nx_print1("Try synching table\n");
	for (i = 0; i < RC_HASH_TBL_SIZE; i++) {
		head = &adapter->rt_table->routes[i];

		spin_lock_bh(&head->lock);

		list_for_each_safe(node1, node2, &head->head) {
			rt = list_entry(node1, nx_host_rt_entry_t, list);
			nx_print1("Found a possible entry\n");
			nx_route_sync(rt, adapter);
		}

		spin_unlock_bh(&head->lock);
	}

	unlock_rt_table(adapter->rt_table);
}

/*
 *
 */
static inline void nx_rt_add_sync_queue(nx_tnic_adapter_t *adapter)
{
	spin_lock_bh(&sync_q.list.lock);

	if (!adapter->rt_sync_scheduled) {
		list_add_tail(&adapter->rt_sync_node, &sync_q.list.head);
		adapter->rt_sync_scheduled = 1;
	}

	spin_unlock_bh(&sync_q.list.lock);

	NX_KC_SCHEDULE_DELAYED_WORK(&sync_q.task, 3 * HZ);
}

/*
 *
 */
static void nx_rt_remove_sync_queue(nx_tnic_adapter_t *adapter)
{
	spin_lock_bh(&sync_q.list.lock);

	if (adapter->rt_sync_scheduled) {
		list_del_init(&adapter->rt_sync_node);
	}

	spin_unlock_bh(&sync_q.list.lock);
}

/*
 *
 */
static void nx_rt_sync_queue_proc(void *data)
{
	nx_tnic_adapter_t	*adapter = NULL;

	spin_lock_bh(&sync_q.list.lock);
	while (!list_empty(&sync_q.list.head)) {
		adapter = list_entry(sync_q.list.head.next, nx_tnic_adapter_t,
				     rt_sync_node);
		list_del_init(&adapter->rt_sync_node);
		adapter->rt_sync_scheduled = 0;
		spin_unlock_bh(&sync_q.list.lock);
		nx_route_sync_table(adapter);
		spin_lock_bh(&sync_q.list.lock);
	}
	spin_unlock_bh(&sync_q.list.lock);
}

/*
 *
 */
void nx_proc_route_invalid(nx_tnic_adapter_t *adapter,
			   routecache_invalid_t *msg)
{
	struct rtable		*rt;
	struct net_device	*dev;
	int			port;
	int			rv;
	struct flowi		fl =
		{ .oif = 0,
		  .nl_u = { .ip4_u = { .daddr = msg->dest.v4,
				       .saddr = msg->src.v4,
				       .tos   = msg->tos } },
		  .proto = msg->protocol,
		  .uli_u = { .ports = { .sport = msg->sport,
					.dport = msg->dport} } };

	rv = ip_route_output_key(&rt, &fl);
	if (rv) {
		nx_print1("Could not find route for saddr[0x%x], daddr[0x%x]: "
			  "ERROR[%d]\n", msg->src.v4, msg->dest.v4, rv);
		return;
	}

        dev = NX_KC_ROUTE_DEVICE(rt);

	/*
	 * Once there is multi card code we need to check if we are going
	 * to the same card.
	 */
	port = nx_get_device_port(dev);
	if (netif_device_present(dev) && is_netxen_device(dev) && port >= 0) {
		nx_route_update(adapter, rt, port);
	}

	ip_rt_put(rt);

	return;
}

/*
 *
 */
int nx_route_connection(struct rtable **rp, uint32_t saddr, uint16_t sport,
			uint32_t daddr, uint16_t dport, uint8_t tos, int oif)
{
	int			rv = 0;
	struct net_device	*dev;
	struct flowi		fl =
		{ .oif = oif,
		  .nl_u = { .ip4_u = { .daddr = daddr,
				       .saddr = saddr,
				       .tos   = tos } },
		  .proto = IPPROTO_TCP,
		  .uli_u = { .ports = { .sport = sport,
					.dport = dport} } };

	dev = netxen_find_dev_for_ip(saddr);
	if (dev) {
		rv = is_netxen_device(dev);
		dev_put(dev);

		if (!rv) {
			return (-1);
		}
	}

	rv = ip_route_output_key(rp, &fl);
	if (rv) {
		return (rv);
	}

	if ((*rp)->rt_flags & (RTCF_MULTICAST | RTCF_BROADCAST)) {
		ip_rt_put(*rp);
		return (-ENETUNREACH);
	}

        dev = NX_KC_ROUTE_DEVICE(*rp);

	if (!netif_device_present(dev) || !is_netxen_device(dev)) {
		ip_rt_put(*rp);
		return (-1);
	}

	return (0);
}


/*
 *
 */
static int parse_rtattr(struct rtattr *tb[], int max, struct rtattr *rta,
			int len)
{
	memset(tb, 0, sizeof(struct rtattr *) * (max + 1));
	while (RTA_OK(rta, len)) {
		if (rta->rta_type <= max)
			tb[rta->rta_type] = rta;
		rta = RTA_NEXT(rta,len);
	}

	if (len) {
		nx_print("!!!Deficit %d, rta_len=%d\n", len, rta->rta_len);
	}
	return 0;
}

/*
 *
 */
static void print_netlink_rt_msg(struct nlmsghdr *n)
{
	struct rtmsg	*r = NLMSG_DATA(n);
	int		len = n->nlmsg_len;
	struct rtattr	*tb[RTA_MAX+1];
	uint32_t	dst;
	uint32_t	src;
	uint32_t	prefsrc;
	uint32_t	via;
	int		iif = 0;
	int		oif = 0;

	len -= NLMSG_LENGTH(sizeof(*r));
	if (len < 0) {
		nx_print("BUG: wrong nlmsg len %d\n", len);
		return;
	}

	if (r->rtm_family != AF_INET) {
		nx_print1("Only IPv4 Route supported\n");
		return;
	}

	nx_print("rtm_family = %u\n", r->rtm_family);
	nx_print("rtm_dst_len = %u\n", r->rtm_dst_len);
	nx_print("rtm_src_len = %u\n", r->rtm_src_len);
	nx_print("rtm_tos = %u\n", r->rtm_tos);
	nx_print("rtm_table = %u\n", r->rtm_table);
	nx_print("rtm_protocol = %u\n", r->rtm_protocol);
	nx_print("rtm_scope = %u\n", r->rtm_scope);
	nx_print("rtm_type = %u\n", r->rtm_type);
	nx_print("rtm_flags = %u\n", r->rtm_flags);

	parse_rtattr(tb, RTA_MAX, RTM_RTA(r), len);

	if (r->rtm_dst_len > 32 || r->rtm_src_len > 32) {
		nx_print("dst/src length > 32\n");
		return;
	}

	if (tb[RTA_DST]) {
		memcpy(&dst, RTA_DATA(tb[RTA_DST]), (r->rtm_dst_len+7)/8);
		nx_print("DST = 0x%x\n", dst);
	}

	if (tb[RTA_SRC]) {
		memcpy(&src, RTA_DATA(tb[RTA_SRC]), (r->rtm_src_len+7)/8);
		nx_print("SRC = 0x%x\n", src);
	}

	if (tb[RTA_GATEWAY]) {
		memcpy(&via, RTA_DATA(tb[RTA_GATEWAY]), 4);
		nx_print("VIA = 0x%x\n", via);
	}

	if (tb[RTA_PREFSRC]) {
		memcpy(&prefsrc, RTA_DATA(tb[RTA_PREFSRC]), 4);
		nx_print("PREFSRC = 0x%x\n", prefsrc);
	}

	if (tb[RTA_IIF]) {
		iif = *(int*)RTA_DATA(tb[RTA_IIF]);
		nx_print("Input interface = %d\n", iif);
	}

	if (tb[RTA_OIF]) {
		oif = *(int*)RTA_DATA(tb[RTA_OIF]);
		nx_print("Output interface = %d\n", oif);
	}

	if (tb[RTA_PRIORITY]) {
		nx_print(" metric %d ", *(__u32*)RTA_DATA(tb[RTA_PRIORITY]));
	}
	if (r->rtm_flags & RTNH_F_DEAD)
		nx_print("dead ");
	if (r->rtm_flags & RTNH_F_ONLINK)
		nx_print("onlink ");
	if (r->rtm_flags & RTNH_F_PERVASIVE)
		nx_print("pervasive ");
	if (r->rtm_flags & RTM_F_EQUALIZE)
		nx_print("equalize ");
	if (r->rtm_flags & RTM_F_NOTIFY)
		nx_print("notify ");
}

/*
 *
 */
static void proc_netlink_rt_msg(struct nlmsghdr *n)
{
	struct rtmsg		*rt = NLMSG_DATA(n);
	int			len = n->nlmsg_len;
	struct rtattr		*attr[RTA_MAX+1];
	int			iif = 0;
	int			oif = 0;
	struct net_device	*dev;
	nx_tnic_adapter_t	*tnic_adapter;

	if (0) {
		print_netlink_rt_msg(n);
	}

	len -= NLMSG_LENGTH(sizeof(*rt));
	if (len < 0) {
		nx_print("BUG: wrong nlmsg len %d\n", len);
		return;
	}

	if (rt->rtm_family != AF_INET) {
		nx_print1("Only IPv4 Route supported\n");
		return;
	}

	parse_rtattr(attr, RTA_MAX, RTM_RTA(rt), len);

	if (attr[RTA_IIF]) {
		iif = *(int*)RTA_DATA(attr[RTA_IIF]);
		nx_print1("Input interface = %d\n", iif);
	}

	if (attr[RTA_OIF]) {
		oif = *(int*)RTA_DATA(attr[RTA_OIF]);
		nx_print1("Output interface = %d\n", oif);
		dev = dev_get_by_index(oif);
		if (!dev) {
			return;
		}
		tnic_adapter = nx_get_dev_tnic_adapter(dev);
		if (tnic_adapter == NULL) {
			nx_print1("Not NetXen device\n");
			dev_put(dev);
			return;
		}
		nx_print1("This is a route affecting our device\n");
		nx_rt_add_sync_queue(tnic_adapter);
		dev_put(dev);
	}
}


#define RCV_SKB_FAIL(err) do { netlink_ack(skb, nlh, (err)); return; } while (0)

static inline void nx_route_rcv_skb(struct sk_buff *skb)
{
	int		type;
	int		pid;
	int		flags;
	int		nlmsglen;
	int		skblen;
        struct nlmsghdr	*nlh;

	nx_print1("Process SKB\n");

        while (skb->len >= NLMSG_SPACE(0)) {

		nx_print1("Process Message\n");

		skblen = skb->len;
		if (skblen < sizeof (*nlh)) {
			nx_print("Bad Length: skblen[%d/%d] < \n",
				 skblen, (int)sizeof(*nlh));
			return;
		}

		nlh = (struct nlmsghdr *)skb->data;
		nlmsglen = nlh->nlmsg_len;
		if (nlmsglen < sizeof(*nlh) || skblen < nlmsglen) {
			nx_print("Bad Length: nlmsglen[%d/%d], skblen[%d]\n",
				 nlmsglen, (int)sizeof(*nlh), skblen);
			return;
		}

		pid = nlh->nlmsg_pid;
		flags = nlh->nlmsg_flags;
        
		if (flags & (NLM_F_MULTI | NLM_F_ACK | NLM_F_ECHO)) {
			nx_print("Flags[0x%x] we are not ready to deal with "
				 "or bad pid[0x%x]\n", flags, pid);
			RCV_SKB_FAIL(-EINVAL);
		}
                
		if (flags & MSG_TRUNC) {
			nx_print("Truncated message: Flags[0x%x]\n", flags);
			RCV_SKB_FAIL(-ECOMM);
		}

		type = nlh->nlmsg_type;
		if (type == RTM_NEWROUTE || type == RTM_DELROUTE) {
			nx_print1("Route NOTICE\n");
			proc_netlink_rt_msg(nlh);
		} else {
			nx_print1("Not route notice\n");
		}

		if (flags & NLM_F_ACK) {
			netlink_ack(skb, nlh, 0);
		}

		skb_pull(skb, NLMSG_ALIGN(nlmsglen));
        }
        return;
}

static struct socket	*nx_rt_sock = NULL;

static void nx_route_data_available(struct sock *sk, int bytes)
{
	struct sk_buff	*skb;

	nx_print1("Received data on route socket\n");

	while ((skb = skb_dequeue(&sk->sk_receive_queue)) != NULL) {
		nx_print1("Dequeued a skb\n");
		nx_route_rcv_skb(skb);
		kfree_skb(skb);
	}
}

static int nx_route_register_listener(struct socket *sock)
{
	struct sockaddr_nl	local;
	int			addr_len;
	int			rv;

	memset(&local, 0, sizeof (local));
	local.nl_family = AF_NETLINK;

#if	LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,14)
	local.nl_groups = 1 << (RTNLGRP_IPV4_ROUTE - 1);
#else
	local.nl_groups = RTMGRP_IPV4_ROUTE;
#endif

	rv = sock->ops->bind(sock, (struct sockaddr *)&local, sizeof (local));
	if (rv < 0) {
		nx_print("Cannot bind netlink socket");
		return (rv);
	}
	addr_len = sizeof(local);
	rv = sock->ops->getname(sock, (struct sockaddr *)&local, &addr_len, 0);
	if (rv < 0) {
		nx_print("Cannot getname\n");
		return (rv);
	}

	if (addr_len != sizeof(local)) {
		nx_print("Wrong address length %d\n", addr_len);
		return (rv);
	}

	if (local.nl_family != AF_NETLINK) {
		nx_print("Wrong address family %d\n", local.nl_family);
		return (rv);
	}

	return (0);
}

/*
 *
 */
static int nx_route_listener_init(void)
{
        struct socket	*sock;
	int		rv;

        if (sock_create_kern(PF_NETLINK, SOCK_RAW, NETLINK_ROUTE, &sock)) {
		nx_print("Socket create failed\n");
                return (-1);
	}

	sock->sk->sk_reuse = 1;
	sock->sk->sk_allocation = GFP_ATOMIC; /* we're going to call it
						 from interrupts */
	sock->sk->sk_data_ready = nx_route_data_available;

	rv = nx_route_register_listener(sock);
	if (rv) {
		sock_release(sock);
		return (-1);
	}

	nx_rt_sock = sock;

	nx_print("Initialized Route Listener Socket\n");
	return (0);
}

/*
 *
 */
static void nx_route_listener_destroy(void)
{
	if (nx_rt_sock != NULL) {
		sock_release(nx_rt_sock);
		nx_rt_sock = NULL;
	}
}
