/*
 * Copyright (C) 2007 NetXen Inc.
 * 
 * All rights reserved.
 * 
 * The code contained in this file has been developed by NetXen Inc.
 * 
 * By downloading, compiling or using the code in any form, you agree to be bound
 * by the following terms:
 * 
 * 1) You may compile the unmodified source code file and reproduce and install
 * the resulting binary code file solely for your internal use with NetXen Inc.
 * hardware and systems you have purchased.
 * 
 * 2) Except as expressly permitted herein or in a separately signed agreement,
 * any other use, reproduction, distribution, modification, display, creation
 * of derivative works or circulation of the code is strictly prohibited.  You
 * may not transfer the code or assign your rights or obligations hereunder
 * without the prior written consent of NetXen Inc.
 * 
 * 3) This file contains confidential, proprietary and trade secret information
 * of NetXen Inc. and may not be disclosed to any third party or used for any
 * other purpose, except as expressly permitted herein.
 * 
 * 4) THIS SOFTWARE IS PROVIDED “AS IS” AND ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL NETXEN INC.
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.  THE APPLICATION OF THE UNITED NATIONS CONVENTION
 * ON CONTRACTS FOR THE INTERNATIONAL SALE OF GOODS IS HEREBY DISCLAIMED.
 * 
 * If you do not agree to the foregoing terms, you may not download, use or
 * compile this code in any form and must promptly delete or destroy all copiesi
 * of the code you have downloaded, compiled or reproduced.
 * 
 * Contact Information:
 * licensing@netxen.com
 * NetXen Inc.
 * 3965 Freedom Circle, Fourth Floor
 * Santa Clara, CA 95054
 * 
 */
#include <linux/autoconf.h>
#include <linux/types.h>
#include <linux/module.h>

#include <asm/uaccess.h>
#include <linux/kernel.h>
#include <linux/spinlock.h>
#include <linux/fs.h>

typedef struct {
	char *fmt;
	long p1;
	long p2;
	long p3;
} LogMsg_t;

#define MAXLOGBUF		2048
LogMsg_t logBuf[MAXLOGBUF+1];
int logTail = 0;
int logHead = 0;

long do_immediate_print = 0;

#if defined(UNM_REDIRECTOR)
static spinlock_t logLock = SPIN_LOCK_UNLOCKED;
#endif

    void
logReset(void)
{
#if defined(UNM_REDIRECTOR)
	unsigned long flags;
	spin_lock_irqsave(&logLock, flags);
#endif
	logTail = logHead = 0;
#if defined(UNM_REDIRECTOR)
	spin_unlock_irqrestore(&logLock, flags);
#endif
}

    void
logAdd(char *fmt, long p1, long p2, long p3)
{
	int myidx;
#if defined(UNM_REDIRECTOR)
	unsigned long flags;

	spin_lock_irqsave(&logLock, flags);
#endif
	if (logTail >= MAXLOGBUF) {
		myidx = logTail = 0;
	} else {
		myidx = logTail++;
		if (logTail >= MAXLOGBUF) {
			logTail = 0;
		}
	}
	if (logTail == logHead) {
		if (++logHead >= MAXLOGBUF) {
			logHead = 0;
		}
	}
#if defined(UNM_REDIRECTOR)
	spin_unlock_irqrestore(&logLock, flags);
#endif
	logBuf[myidx].fmt = fmt;
	logBuf[myidx].p1 = p1;
	logBuf[myidx].p2 = p2;
	logBuf[myidx].p3 = p3;
	if (do_immediate_print) {
#if defined(UNM_REDIRECTOR)
		printk(fmt, p1, p2, p3);
#else
		printf(fmt, p1, p2, p3);
#endif
	}
}

    void
logDump(void)
{
	int head;
	int tail;
#if defined(UNM_REDIRECTOR)
	unsigned long flags;

	printk
#else
	printf
#endif
		("logDump : head = %d, tail = %d\n", logHead, logTail);
#if defined(UNM_REDIRECTOR)
	spin_lock_irqsave(&logLock, flags);
#endif
	head = logHead;
	tail = logTail;
#if defined(UNM_REDIRECTOR)
	spin_unlock_irqrestore(&logLock, flags);
#endif
	while (head != tail) {
#if defined(UNM_REDIRECTOR)
		printk
#else
		printf
#endif
			(logBuf[head].fmt, logBuf[head].p1, logBuf[head].p2,
				logBuf[head].p3);
		if (++head >= MAXLOGBUF) {
			head = 0;
		}
	}
#if defined(UNM_REDIRECTOR)
	printk
#else
	printf
#endif
		("Done.\n");
}

#if defined(UNM_REDIRECTOR)
int readHead = 0;
int readTail = 0;
int openCount = 0;

extern void show_state(void);

static int
unm_net_open(struct inode *inode, struct file *file)
{
	unsigned long flags;

	spin_lock_irqsave(&logLock, flags);
	if (openCount > 0) {
		spin_unlock_irqrestore(&logLock, flags);
		return -EBUSY;
	}
	++openCount;
	readHead = logHead;
	readTail = logTail;
	spin_unlock_irqrestore(&logLock, flags);
	// show_state();
	return 0;
}

static ssize_t unm_net_write(struct file *file, const char *buf, size_t sz,
				loff_t *off)
{
	logReset();
	return sz;
}

static ssize_t unm_net_read(struct file *file, char *buf, size_t size, loff_t *off)
{
	char kbuf[256];
	int len;
	ssize_t count = 0;

	while (readHead != readTail) {
		len = snprintf(kbuf, sizeof(kbuf)-1, logBuf[readHead].fmt,
				logBuf[readHead].p1,
				logBuf[readHead].p2, logBuf[readHead].p3);
		if (len + count > size) {
			*off += count;
			return count;
		}
		count += len;
		if (copy_to_user(buf, kbuf, len)) {
			return -EFAULT;
		}
		buf += len;
		if (++readHead >= MAXLOGBUF) {
			readHead = 0;
		}
	}
	*off += count;
	return count;
	return 0;
}

static int unm_net_release(struct inode *inode, struct file *file)
{
	unsigned long flags;

	spin_lock_irqsave(&logLock, flags);
	--openCount;
	spin_unlock_irqrestore(&logLock, flags);
	return 0;
}

static struct file_operations unm_net_fops = {
	read: unm_net_read,
	write: unm_net_write,
	open: unm_net_open,
	release: unm_net_release,
};

int unm_net_major = 0;

int unm_log_init(void)
{
   if ((unm_net_major = register_chrdev(unm_net_major, "unm_net", &unm_net_fops)) < 0) {
	printk("Could not allocate major device for unm_net\n");
   } else {
	printk("UNM: allocated major device number %d for unm_net\n", unm_net_major);
   }
   return 0;
}

void unm_log_exit(void)
{
    unregister_chrdev(unm_net_major, "unm_net");
}

#if !defined(NO_MODULE)
#if (LINUX_VERSION_CODE < 0x02032a)
#ifdef MODULE
int init_module(void)
{
	return unm_log_init();
}

void cleanup_module()
{
	unm_log_exit();
}
#endif
#else
module_init(unm_log_init);
module_exit(unm_log_exit);
#endif
EXPORT_SYMBOL(logAdd);
EXPORT_SYMBOL(logDump);
EXPORT_SYMBOL(logReset);

MODULE_AUTHOR("Universal Network Machines");
MODULE_DESCRIPTION("A Simple log driver for debugging help");
MODULE_LICENSE("GPL");
#endif /* NO_MODULE */
#endif /* UNM_REDIRECTOR */
