/*
 * Copyright (C) 2007 NetXen Inc.
 * 
 * All rights reserved.
 * 
 * The code contained in this file has been developed by NetXen Inc.
 * 
 * By downloading, compiling or using the code in any form, you agree to be bound
 * by the following terms:
 * 
 * 1) You may compile the unmodified source code file and reproduce and install
 * the resulting binary code file solely for your internal use with NetXen Inc.
 * hardware and systems you have purchased.
 * 
 * 2) Except as expressly permitted herein or in a separately signed agreement,
 * any other use, reproduction, distribution, modification, display, creation
 * of derivative works or circulation of the code is strictly prohibited.  You
 * may not transfer the code or assign your rights or obligations hereunder
 * without the prior written consent of NetXen Inc.
 * 
 * 3) This file contains confidential, proprietary and trade secret information
 * of NetXen Inc. and may not be disclosed to any third party or used for any
 * other purpose, except as expressly permitted herein.
 * 
 * 4) THIS SOFTWARE IS PROVIDED “AS IS” AND ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL NETXEN INC.
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.  THE APPLICATION OF THE UNITED NATIONS CONVENTION
 * ON CONTRACTS FOR THE INTERNATIONAL SALE OF GOODS IS HEREBY DISCLAIMED.
 * 
 * If you do not agree to the foregoing terms, you may not download, use or
 * compile this code in any form and must promptly delete or destroy all copiesi
 * of the code you have downloaded, compiled or reproduced.
 * 
 * Contact Information:
 * licensing@netxen.com
 * NetXen Inc.
 * 3965 Freedom Circle, Fourth Floor
 * Santa Clara, CA 95054
 * 
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>    
#include <sys/types.h>          
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <netinet/in.h>
#include "netxen_linux_ioctl.h"


static void usage(const char*);
static void parse_tuple(char *str, netxen_offload_rule_t *rule);
static int list_rules(int sd, netxen_offload_ioctl_t *data);

static char     *prog_name;

int main(int argc, char *argv[])
{
        netxen_offload_ioctl_t  data;
        int                     rule_opt_cnt = 0;
        int                     cmd_opt_cnt = 0;
        char                    err_str[256];
	int                     c;
        int                     sd;
        int                     rv;
        int                     int_val;

        prog_name = argv[0];
	if (argc < 2) {
		usage(NULL);
        }

        memset(&data, 0, sizeof (data));

	while ((c = getopt(argc, argv, "?hp:n:ardszt:")) != EOF) {
		switch (c) {
                        /* COMMANDS */
                case 'a':
                        data.hdr.cmd = NETXEN_OFFLOAD_IOCTL_ADD_RULE;
                        cmd_opt_cnt++;
                        break;

                case 'r':
                        data.hdr.cmd = NETXEN_OFFLOAD_IOCTL_REMOVE_RULE;
                        cmd_opt_cnt++;
                        break;

                case 's':
                        data.hdr.cmd = NETXEN_OFFLOAD_IOCTL_GET_RULES;
                        cmd_opt_cnt++;
                        break;

                        /* RULES */
                case 'z':
                        data.u.rule.zero_copy = 1;
                        rule_opt_cnt++;
                        break;

                case 'p':
                        data.u.rule.type = NETXEN_OFFLOAD_RULE_BY_TUPLE;
                        int_val = atoi(optarg);
                        if (int_val <= 0 || int_val >= 65536) {
                                usage("Invalid port: Valid range 1 to 65535");
                        }
                        data.u.rule.local_port = int_val;
                        rule_opt_cnt++;
                        break;

                case 'n':
                        data.u.rule.type = NETXEN_OFFLOAD_RULE_BY_APP_NAME;
                        if (strlen(optarg) > 49) {
                                usage("Length of app name should be less than "
                                      "50 characters");
                        }
                        strcpy(data.u.rule.app_name, optarg);
                        rule_opt_cnt++;
                        break;

		case 't':
                        data.u.rule.type = NETXEN_OFFLOAD_RULE_BY_TUPLE;
			parse_tuple(optarg, &data.u.rule);
                        rule_opt_cnt++;
			break;

		case 'h':
		case '?':
			usage(NULL);

                default:
                        sprintf(err_str, "Unrecognized option %c", (char)c);
                        usage(err_str);
		}
	}

        if (cmd_opt_cnt != 1) {
                if (cmd_opt_cnt == 0) {
                        usage("One of the options [a,r,s,d,c] has to be "
                              "specified");
                } else {
                        usage("Only one of the options [a,r,s,d,c] can be "
                              "specified");
                }
        }

        if (data.hdr.cmd == NETXEN_OFFLOAD_IOCTL_ADD_RULE ||
            data.hdr.cmd == NETXEN_OFFLOAD_IOCTL_REMOVE_RULE) {
                if (rule_opt_cnt != 1 && rule_opt_cnt != 2) {
                        if (rule_opt_cnt == 0) {
                                usage("One of the rules [p,n] and optional [z] has to be "
                                      "specified");
                        } else {
                                usage("Only one of the rules [p,n] can be "
                                      "specified");
                        }
                }

        } else if (data.hdr.cmd == NETXEN_OFFLOAD_IOCTL_GET_RULES) {
                if (rule_opt_cnt) {
                        usage("Rules cannot be specified with options [s,d,c]");
                }
        }

	/*create socket*/
	sd = socket(AF_INET, SOCK_DGRAM, 0);			
	if (sd < 0) {
		printf("Unable to create socket.\n");
		exit (1);
	}

        switch (data.hdr.cmd) {
        case NETXEN_OFFLOAD_IOCTL_ADD_RULE:
        case NETXEN_OFFLOAD_IOCTL_REMOVE_RULE:
                rv = ioctl(sd, SIOC_NETXEN_OFFLOAD, (char *)&data);
                if (rv == 0) {
                        if (data.hdr.ret_code != 0) {
                                printf("Error[%d]: Request failed\n",
                                       (int)data.hdr.ret_code);
                        }
                }
                break;

        case NETXEN_OFFLOAD_IOCTL_GET_RULES:
                rv = list_rules(sd, &data);
                break;
        }

        close(sd);

        if (rv < 0) {
		switch ((int)data.hdr.ret_code) {
		case -NETXEN_IOCTL_ERR_RULE_EXISTS:
			printf("Rule Already Exists\n");
			break;

		case -NETXEN_IOCTL_ERR_RULE_NOT_FOUND:
			printf("Rule Not found\n");
			break;

		case -NETXEN_IOCTL_ERR_UNSUPPORTED_RULE:
			printf("Unsupported Rule\n");
			break;

		default:
		{
			uid_t	id = geteuid();

			if (id > 0 &&
			    (data.hdr.cmd == NETXEN_OFFLOAD_IOCTL_ADD_RULE ||
			     data.hdr.cmd == NETXEN_OFFLOAD_IOCTL_REMOVE_RULE)) {
				printf("To add or delete a rule superuser "
				       "privilege is required.\n");
			} else {
				printf("nxoffload failed \n");
			}
		}
		}
	}
	return (0);
}

static void usage(const char *message)
{
        if (message != NULL) {
                printf("\nError: %s.\n\n", message);
        }

	printf("usage: %s <options>\n", prog_name);
	printf("options: -p   Port number to be offloaded.\n");
	printf("         -n   Application name to be offloaded.\n");
	printf("         -t   TCP tuple to be offloaded.\n");
	printf("              Format: local IP, local port, remote IP, remote port.\n");
	printf("              example:\n");
	printf("                %s -a -t'0, 0, 0, 5001'.\n", prog_name);
	printf("                %s -a -t'0, 2000, 0, 5001'.\n", prog_name);
	printf("         -a   Add an offload rule.\n");
	printf("         -r   Remove an offload rule.\n");
	printf("         -z   Enable zero copy.\n");
	printf("         -s   List all ports and applications offloaded.\n");

	exit (1);
}

/*
 *
 */
static void parse_tuple(char *str, netxen_offload_rule_t *rule)
{
	int		int_val;
	char		*ptr;
	char		*tmp;


	/*
	 * Parse the local IP address.
	 */
	ptr = strtok_r(str, ",", &tmp);
	if (ptr == NULL) {
		printf("Local IP parse failed\n");
		exit(0);
	}

/* 	if (parse_ip(str, &local_ip)) { */
/* 		usage("Invalid Tuple: Sample is " */
/* 		      "'192.150.2.1,80,192.111.2.1,80'"); */
/* 	} */

	/*
	 * Parse local port.
	 */
	ptr = strtok_r(NULL, ",", &tmp);
	if (ptr == NULL) {
		printf("Local port parse failed\n");
		exit(0);
	}

	int_val = atoi(ptr);
	if (int_val < 0 || int_val >= 65536) {
		usage("Invalid port: Valid range 0 to 65535");
	}
	rule->local_port = int_val;

	/*
	 * Parse remote IP.
	 */
	ptr = strtok_r(NULL, ",", &tmp);
	if (ptr == NULL) {
		printf("Remote IP parse failed\n");
		exit(0);
	}

	/*
	 * Parse remote port.
	 */
	ptr = strtok_r(NULL, ",", &tmp);
	if (ptr == NULL) {
		printf("Remote port parse failed\n");
		exit(0);
	}

	int_val = atoi(ptr);
	if (int_val < 0 || int_val >= 65536) {
		usage("Invalid port: Valid range 0 to 65535");
	}
	rule->remote_port = int_val;
}

/*
 *
 */
static int list_rules(int sd, netxen_offload_ioctl_t *data)
{
        int                             rv;
        char                            *buffer;
        int                             i;
        netxen_offload_ioctl_hdr_t      *hdr;
        netxen_offload_rule_t           *rule;
        uint16_t                        cnt;


        data->hdr.cmd = NETXEN_OFFLOAD_IOCTL_GET_RULE_CNT;
        rv = ioctl(sd, SIOC_NETXEN_OFFLOAD, (char *)data);
        if (rv < 0) {
                printf("Error[%d]: Ioctl failed.\n", rv);
                return rv;
        }
        if (data->hdr.cnt == 0) {
                printf("\nNo offload rules configured.\n");
                return rv;
        } else {
                char    str[256];
                int     len;

                len = sprintf(str, "%u offload rule%s configured.",
                              data->hdr.cnt,
                              ((data->hdr.cnt == 1) ? "" : "s"));
                printf("\n%16s%s\n%16s", " ", str, " ");
                while (len > 0) {
                        printf("=");
                        len--;
                }
                printf("\n\n");
        }

        cnt = data->hdr.cnt;
        buffer = malloc(sizeof (netxen_offload_ioctl_hdr_t) +
                        (cnt * sizeof (netxen_offload_rule_t)));
        if (buffer == NULL) {
                printf("Error: Allocation to list %d buffers failed. Try a "
                       "smaller count.\n", cnt);
                return (0);
        }

        data->hdr.cmd = NETXEN_OFFLOAD_IOCTL_GET_RULES;
        memcpy(buffer, data, sizeof (netxen_offload_ioctl_hdr_t));

        rv = ioctl(sd, SIOC_NETXEN_OFFLOAD, buffer);
        if (rv < 0) {
                goto done;
        }

        hdr = (netxen_offload_ioctl_hdr_t *)buffer;
        rv = (int)hdr->ret_code;
        if (rv < 0) {
                printf("Error[%d]: Request for rule list failed\n", rv);
                rv = 0;
                goto done;
        }

        if (hdr->cnt > cnt) {
                printf("Error: Returned count[%d] greater than requested "
                       "count[%d].\n", hdr->cnt, cnt);
                rv = 0;
                goto done;
        }

        cnt = hdr->cnt;
        rule = (netxen_offload_rule_t *)(buffer +
                                         sizeof (netxen_offload_ioctl_hdr_t));
        for (i = 0; i < cnt; i++) {
                switch (rule[i].type) {
                case NETXEN_OFFLOAD_RULE_BY_TUPLE:
                        printf("%8s%d. Local Port - Remote Port        :   %u - %u"
                               " (TX zero copy %s)\n", " ",
                               i + 1, rule[i].local_port, rule[i].remote_port,
                               rule[i].zero_copy? "on" : "off");
                        break;

                case NETXEN_OFFLOAD_RULE_BY_APP_NAME:
                        printf("%8s%d. Application Name                :   %s"
                               " (TX zero copy %s)\n", " ",
                               i + 1, rule[i].app_name,
                               rule[i].zero_copy? "on" : "off");
                        break;

                default:
                        printf("%d. Unknown rule type[%u]\n", i + 1,
                               rule[i].type);
                        break;
                }
        }

        printf("\n\n");
        if (rv) {
                printf("There are more rules[%d] in the system.\n\n", rv);
                rv = 0;
        }

  done:
        free(buffer);
        return (rv);
}
